/*
 * This file is part of the Arduino I2C keyboard.
 * Copyright (C) 2018  Forest Godfrey <fgodfrey@bigw.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


#ifndef _AI2C_INTERFACE_H
#define _AI2C_INTERFACE_H

/*
 * Address on the I2C Bus
 */
#define AI2C_DEVICE_ADDRESS     0x42

/*
 * The basic interface is as follows:
 * The device generates an interrupt on any of the following conditions:
 * - Key pressed
 * - Key released
 * - Camera operation finished
 * - Battery status change
 * 
 * To write data, the host sends a write message with three 8 bit words:
 * <i2c address> <register address> <register data>
 * 
 * To read data, the host writes AI2C_REGISTER_READ_ADDR and 
 * AI2C_REGISTER_READ_BYTES.  The host can then repeatedly read the device.
 * Each read will produce the amount of data requested starting at the written
 * address.  For instance, the host could set the read address to 
 * AI2C_REGISTER_KEY_DOWN_LOW and the length to 6 bytes and read all 6 bytes 
 * of keyboard status every time a read is generated.
 * 
 * The following abreviations are used for each register:
 * - RO:  Read Only - The register is managed by the Arduino.  Writes will be silently dropped
 * - RC:  Read Clear - Reading this register clears its contents
 * - RW:  Read Write - This register can be read or written.
 * - RWC:  Read Write/Clear - Register can be read.  Write 1 to a bit to clear it.  Individual bits may be RO
 */

/*
 * AI2C Registers
 */
#define AI2C_REGISTER_ID_LOW             0x00    /* (RO) Device ID Low 8 bits */
#define AI2C_REGISTER_ID_HIGH            0x01    /* (RO) Device ID High 8 bits */
#define AI2C_REGISTER_REG_REV            0x02    /* (RO) Register interface compatibility version */
#define AI2C_REGISTER_VERS_STR           0x03    /* (RO) Version string - 64 bytes */
#define AI2C_REGISTER_KEY_STATUS_LOW     0x44    /* (RO) Keys down/up */
#define AI2C_REGISTER_KEY_STATUS_HIGH    0x45
#define AI2C_REGISTER_KEY_DOWN_LOW       0x46    /* (RC) Keys pressed since last check - Read-Clear */
#define AI2C_REGISTER_KEY_DOWN_HIGH      0x47
#define AI2C_REGISTER_KEY_UP_LOW         0x48    /* (RC) Keys released since last check - Read-Clear */
#define AI2C_REGISTER_KEY_UP_HIGH        0x49
#define AI2C_REGISTER_CAMERA_FRAC_LOW    0x4a    /* (RW) Fractional Camera Shutter Speed in 100's of uSecs */
#define AI2C_REGISTER_CAMERA_FRAC_HIGH   0x4b
#define AI2C_REGISTER_CAMERA_SECS        0x4c    /* (RW) Shutter Speed in Seconds - added to fraction */
#define AI2C_REGISTER_COMMAND            0x4d    /* (RW) Execute command (like "focus" or "fire") */
#define AI2C_REGISTER_READ_ADDR          0x4e    /* (RW) Address to read on a read command */
#define AI2C_REGISTER_READ_BYTES         0x4f    /* (RW) Number of bytes to receive on a read */
#define AI2C_REGISTER_STATUS             0x50    /* (RWC) Various status bits */
#define AI2C_INTR_MASK_LOW               0x51    /* (RW) Interrupt mask */
#define AI2C_INTR_MASK_HIGH              0x52
#define AI2C_POWER_STATUS                0x53    /* (RO) Battery status */
#define AI2C_BAT_VOLTAGE_LOW             0x54    /* (RO) Battery voltage XXX - units? */
#define AI2C_BAT_VOLTAGE_HIGH            0x55
#define AI2C_REGISTER_MAX                0x53    /* Must be set to the max register address */

/*
 * Commands
 */
#define AI2C_CMD_CMD_COMPLETE            0x00    /* Previous command completed - must be read back */

/*
 * Device ID
 */
#define AI2C_ID                          0x4242

/*
 * Register revision ID.  Change when register interface becomes incompatible
 */
#define AI2C_REV                         0x01

/*
 * Max length of build version string.  Registers will be padded with zeros
 */
#define AI2C_VER_STR_LEN                 64
#endif /* _AI2C_INTERFACE_H */
