/*
 * Arduino I2C Keyboard
 * This kernel module provides evdev events for keys controlled by an I2C device.  An example
 * compatible device is provided as an Arduino sketch in the directory above.
 *
 * Copyright (C) 2018  Forest Godfrey <fgodfrey@bigw.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/*
 * Kernel Includes
 */
#include <linux/module.h>

#include <linux/gpio.h>
#include <linux/i2c.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/workqueue.h>

/*
 * Project Includes
 */
#include "ai2c.h"
#include "ai2c_platform_data.h"
#include "ai2c_interface.h"

#define IRQ_GPIO_PIN	9
#define GPIO_IRQ_NAME	"ai2c_keyboard_irq"

static int default_keymap[] = {
	KEY_UP,
	KEY_DOWN,
	KEY_LEFT,
	KEY_RIGHT,
	KEY_A,
	KEY_B,
	BTN_START,
	KEY_SELECT,
	KEY_DISPLAYTOGGLE,
	KEY_1,
	KEY_2,
	KEY_3,
	KEY_4,
	KEY_5,
	KEY_6,
	KEY_7,
	KEY_8,
	KEY_9
};

static int default_numkeys = sizeof(default_keymap) / sizeof(int);

static LIST_HEAD(driver_list);
static DEFINE_MUTEX(driver_list_lock);

static void
device_write_register(struct i2c_client *client, uint8_t raddr, uint8_t value) {
	uint8_t tmp[2];

	tmp[0] = raddr;
	tmp[1] = value;
	i2c_master_send(client, tmp, 2);
}

#if 0
/* ifdef'd out to pacify gcc until we need there */
static uint8_t
device_read_byte(struct i2c_client *client) {
	uint8_t tmp;

	i2c_master_recv(client, &tmp, 1);

	return tmp;
}

static uint16_t
device_read_word(struct i2c_client *client) {
	uint16_t tmp;

	i2c_master_recv(client, (unsigned char *)&tmp, 1);

	return tmp;
}
#endif

static void
device_read_register(struct i2c_client *client, void *buffer, uint8_t raddr, uint8_t len) {
	device_write_register(client, AI2C_REGISTER_READ_ADDR, raddr);
	device_write_register(client, AI2C_REGISTER_READ_BYTES, len);
	i2c_master_recv(client, (unsigned char *)buffer, len);
}

static void
ai2c_intr(struct work_struct *work) {
	int i;
	uint8_t key_status[6];
	uint16_t cur, down, up;
	arduino_i2c_driver_data_t *drv_info = container_of(
			work, arduino_i2c_driver_data_t, intr_upper_work);

	BUG_ON(!drv_info->client);
	dev_info(&drv_info->client->dev, "ai2c_intr:  entered\n");

	memset(key_status, 0, sizeof(key_status));
	for (i = 0; i < 6; i++) {
		device_read_register(drv_info->client, &(key_status[i]), AI2C_REGISTER_KEY_STATUS_LOW + i, 1);
	}
	cur = (uint16_t)key_status[0] | ((uint16_t)key_status[1] << 8);
	down = (uint16_t)key_status[2] | ((uint16_t)key_status[3] << 8);
	up = (uint16_t)key_status[4] | ((uint16_t)key_status[5] << 8);
	for (i = 0; i < 16; i++) {
		if ((down & (1 << i)) && (up & (1 << i))) {
			/* Key went down and up since last check - see what it is currently to decide order */
			if (cur & (1 << i)) {
				/* The button is currently up so it must have gone down first, then up */
				input_report_key(drv_info->indev, default_keymap[i], 0);
				input_sync(drv_info->indev);
				input_report_key(drv_info->indev, default_keymap[i], 1);
			} else {
				/* Other way agound - up first, then down */
				input_report_key(drv_info->indev, default_keymap[i], 1);
				input_sync(drv_info->indev);
				input_report_key(drv_info->indev, default_keymap[i], 0);
			}
		} else if (down & (1 << i)) {
			/* New key down */
			input_report_key(drv_info->indev, default_keymap[i], 0);
		} else if (up & (1 << i)) {
			/* New key up */
			input_report_key(drv_info->indev, default_keymap[i], 1);
		}
	}
	/* All events before this happened at the "same" time */
	input_sync(drv_info->indev);

	return;
}

static irqreturn_t
ai2c_intr_lower(int irq, void *data) {
	arduino_i2c_driver_data_t *drv_info = (arduino_i2c_driver_data_t *)data;

	BUG_ON(!drv_info);
	if (drv_info->irq_count++ < 10) {
		printk("ai2c_intr_lower:  entered\n");
	}
	schedule_work(&drv_info->intr_upper_work);
	return IRQ_HANDLED;
}

static int
ai2c_probe(struct i2c_client *client, const struct i2c_device_id *id) {
	int i, keymax;
	int ret = 0;
	int flags = 0;
	uint16_t dev_id;
	uint8_t dev_rev;
	char dev_build_str[AI2C_VER_STR_LEN];
	const arduino_i2c_keyboard_platform_data_t *pdata =
		dev_get_platdata(&client->dev);
	arduino_i2c_driver_data_t *drv_info = NULL;

	dev_info(&client->dev, "Arduino I2C Keyboard Probing Device");
	if (!pdata) {
		dev_err(&client->dev, "No platform present - failing");
		ret = -EINVAL;
		goto out;
	}

	/*
	 * Get/validate/print device information
	 */
	device_read_register(client, &dev_id, AI2C_REGISTER_ID_LOW, 2);
	device_read_register(client, &dev_rev, AI2C_REGISTER_REG_REV, 1);
	for (i = 0; i < AI2C_VER_STR_LEN; i++) {
		device_read_register(client, &(dev_build_str[i]), AI2C_REGISTER_VERS_STR + i, 1);
	}
	dev_build_str[sizeof(dev_build_str) - 1] = '\0';

	dev_info(&client->dev, "Device ID: 0x%04x\nDevice Revision: 0x%02x\nDevice Firmware Build: \"%s\"\n",
			dev_id, dev_rev, dev_build_str);

	/*
	 * Allocate/initialize driver data
	 */
	drv_info = (arduino_i2c_driver_data_t *)kmalloc(sizeof(arduino_i2c_driver_data_t), GFP_KERNEL);
	if (!drv_info) {
		dev_err(&client->dev, "Unable to allocate driver memory");
		ret = -ENOMEM;
		goto out;
	}
	memset(drv_info, 0, sizeof(arduino_i2c_driver_data_t));

	/*
	 * Get GPIO pin for interrupts
	 */
	ret = gpio_request(pdata->gpio_intr_pin, GPIO_IRQ_NAME);
	if (ret < 0) {
		dev_err(&client->dev, "ai2c:  IRQ pin reservation of %d failed:  error %d", pdata->gpio_intr_pin, ret);
		goto out_dealloc;
	}

	gpio_direction_input(pdata->gpio_intr_pin);
	gpio_set_debounce(pdata->gpio_intr_pin, 50);	// 50ms

#ifndef CONFIG_SIERRA
	/* GPIO exports are broken on Sierra Wireless Linux due to their weird gpiolib mods */
	gpio_export(pdata->gpio_intr_pin, "arduino_i2c_intr_pin");
#endif /* CONFIG_SIERRA */

	drv_info->irq_pin = pdata->gpio_intr_pin;
	drv_info->irq_num = gpio_to_irq(pdata->gpio_intr_pin);

	dev_dbg(&client->dev, "ai2c:  Reserved pin [%d]:  irq = %d\n", pdata->gpio_intr_pin, drv_info->irq_num);

	/*
	 * Register for key press IRQ
	 */
	flags |= IRQF_TRIGGER_FALLING;
	ret = request_irq(drv_info->irq_num, ai2c_intr_lower, flags, "ai2c_keyboard", drv_info);
	if (ret < 0) {
		dev_err(&client->dev, "ai2c:  request_irq failed:  error %d", ret);
		goto out_cleanup;
	}

	drv_info->indev = devm_input_allocate_device(&client->dev);
	if (!drv_info->indev) {
		ret = -ENOMEM;
		dev_err(&client->dev, "ai2c:  unable to allocate input device");
			goto out_cleanirq;
	}

	/*
	 * Set up input device
	 */
	if (pdata->keymap) {
		drv_info->keymap = pdata->keymap;
		drv_info->numkeys = pdata->numkeys;
	} else {
		drv_info->keymap = default_keymap;
		drv_info->numkeys = default_numkeys;
	}
	drv_info->indev->name = "arduino_i2c";
	drv_info->indev->phys = "arduino_i2c/input0";
	drv_info->indev->dev.parent = &client->dev;
	drv_info->indev->id.bustype = BUS_I2C;
	drv_info->indev->id.vendor = AI2C_ID;
	drv_info->indev->id.product = AI2C_ID;
	drv_info->indev->id.version = 0x0001;
	drv_info->indev->keycodesize = sizeof(unsigned short);
	drv_info->indev->evbit[0] = BIT_MASK(EV_KEY);
	keymax = -1;
	for (i = 0; i < drv_info->numkeys; i++) {
		drv_info->indev->keybit[BIT_WORD(drv_info->keymap[i])] |= BIT_MASK(drv_info->keymap[i]);
		keymax = (drv_info->keymap[i] > keymax) ? drv_info->keymap[i] : keymax;
	}
	drv_info->indev->keycodemax = keymax + 1;

	ret = input_register_device(drv_info->indev);
	if (ret) {
		dev_err(&client->dev, "ai2c:  unable to register input device: %d", ret);
		goto out_cleanindev;
	}

	irq_set_irq_type(drv_info->irq_num, IRQ_TYPE_EDGE_FALLING);

	drv_info->client = client;

	INIT_WORK(&drv_info->intr_upper_work, ai2c_intr);
	dev_set_drvdata(&client->dev, drv_info);

	INIT_LIST_HEAD(&drv_info->driver_list);
	mutex_lock(&driver_list_lock);
	list_add(&drv_info->driver_list, &driver_list);
	mutex_unlock(&driver_list_lock);

	return 0;

out_cleanindev:
	input_free_device(drv_info->indev);
out_cleanirq:
	free_irq(drv_info->irq_num, drv_info);
out_cleanup:
	gpio_free(drv_info->irq_pin);
out_dealloc:
	kfree(drv_info);
out:
	return ret;
}

static int
ai2c_remove(struct i2c_client *client) {
	arduino_i2c_driver_data_t *drv_info = (arduino_i2c_driver_data_t *)dev_get_drvdata(&client->dev);
	BUG_ON(!drv_info);

	free_irq(drv_info->irq_num, drv_info);

	flush_work(&drv_info->intr_upper_work);

	gpio_free(drv_info->irq_pin);

	dev_set_drvdata(&drv_info->client->dev, NULL);

	input_unregister_device(drv_info->indev);

	mutex_lock(&driver_list_lock);
	list_del(&drv_info->driver_list);
	mutex_unlock(&driver_list_lock);
	kfree(drv_info);

	return 0;
}

/*
 * Module initialization and parameters.
 */
static const struct i2c_device_id ai2c_id[] = {
	{ "ai2c-keys", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, ai2c_id);

static struct i2c_driver ai2c_driver = {
	.driver = {
		.name = KBUILD_MODNAME,
	},
	.probe    = ai2c_probe,
	.remove   = ai2c_remove,
	.id_table = ai2c_id,
};

static int
ai2c_init(void) {
	int ret;
	printk(KERN_INFO "Arduino I2C Keyboard Starting\n");
	ret = i2c_add_driver(&ai2c_driver);
	if (ret < 0) {
		printk(KERN_ERR "Unable to register AI2C keyboard driver\n");
	}
	return ret;
}

static void
ai2c_shutdown(void) {
	printk(KERN_INFO "Arduino I2C Keyboard Shutting Down\n");

	i2c_del_driver(&ai2c_driver);
}

module_init(ai2c_init);
module_exit(ai2c_shutdown);

MODULE_ALIAS("arduino_i2c");
MODULE_ALIAS("ai2c-keys");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Forest Godfrey <fgodfrey@bigw.org>");
MODULE_DESCRIPTION("evdev driver for i2c keyboard based on Arduino");
MODULE_VERSION("0.1");
